@@ -1,13 +1,34 @@
 #include <EXTERN.h>
-#include <XSUB.h>
 #include <perl.h>
+#include <XSUB.h>
 
 static void xs_init (pTHX);
 EXTERN_C void boot_DynaLoader (pTHX_ CV* cv);
 EXTERN_C void boot_Socket (pTHX_ CV* cv);
 
+typedef struct {
+    I32 key; /* to make sure it came from Inline */
+    IV index;
+    union {
+        SV *(*call_p6_method)(IV, char *, I32, SV *, SV **);
+        SV *(*call_p6_callable)(IV, SV *, SV **);
+    };
+    void (*free_p6_object)(IV);
+} _perl6_magic;
+
+typedef struct {
+    I32 key; /* to make sure it came from Inline */
+    IV index;
+    SV *(*call_p6_method)(IV, char *, I32, SV *, SV **);
+    SV *(*hash_at_key)(IV, char *);
+    SV *(*hash_assign_key)(IV, char *, SV *);
+    void (*free_p6_object)(IV);
+} _perl6_hash_magic;
+
 XS(p5_call_p6_method);
 XS(p5_call_p6_callable);
+XS(p5_hash_at_key);
+XS(p5_hash_assign_key);
 XS(p5_load_module);
 XS(p5_set_subname);
 
@@ -16,6 +37,8 @@ EXTERN_C void xs_init(pTHX) {
     /* DynaLoader is a special case */
     newXS("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
     newXS("Perl6::Object::call_method", p5_call_p6_method, file);
+    newXS("Perl6::Hash::FETCH", p5_hash_at_key, file);
+    newXS("Perl6::Hash::STORE", p5_hash_assign_key, file);
     newXS("Perl6::Callable::call", p5_call_p6_callable, file);
     newXS("v6::load_module_impl", p5_load_module, file);
     newXS("v6::set_subname", p5_set_subname, file);
@@ -25,13 +48,23 @@ size_t p5_size_of_iv() {
     return IVSIZE;
 }
 
+#if NVSIZE > 8
+#    define MYNVSIZE 8
+#    define MYNV double
+#else
+#    define MYNVSIZE NVSIZE
+#    define MYNV NV
+#endif
+
 size_t p5_size_of_nv() {
-    return NVSIZE;
+    return MYNVSIZE;
 }
 
 void p5_inline_perl6_xs_init(PerlInterpreter *my_perl) {
     char *file = __FILE__;
     newXS("Perl6::Object::call_method", p5_call_p6_method, file);
+    newXS("Perl6::Hash::FETCH", p5_hash_at_key, file);
+    newXS("Perl6::Hash::STORE", p5_hash_assign_key, file);
     newXS("Perl6::Callable::call", p5_call_p6_callable, file);
     newXS("v6::load_module_impl", p5_load_module, file);
     newXS("v6::set_subname", p5_set_subname, file);
@@ -41,10 +74,7 @@ static int inited = 0;
 static int interpreters = 0;
 static int terminate = 0;
 
-PerlInterpreter *p5_init_perl() {
-    char *embedding[] = { "", "-e", "0" };
-    int argc = 0;
-    char **argv;
+PerlInterpreter *p5_init_perl(int argc, char **argv) {
     if (inited) {
 #ifndef MULTIPLICITY
         return NULL;
@@ -61,7 +91,7 @@ PerlInterpreter *p5_init_perl() {
     PERL_SET_CONTEXT(my_perl);
     PL_perl_destruct_level = 1;
     perl_construct( my_perl );
-    perl_parse(my_perl, xs_init, 3, embedding, NULL);
+    perl_parse(my_perl, xs_init, argc, argv, NULL);
     PL_exit_flags |= PERL_EXIT_DESTRUCT_END;
     perl_run(my_perl);
 
@@ -70,6 +100,11 @@ PerlInterpreter *p5_init_perl() {
 
 void p5_destruct_perl(PerlInterpreter *my_perl) {
     PL_perl_destruct_level = 1;
+
+    POPSTACK_TO(PL_mainstack);
+    dounwind(-1);
+    LEAVE_SCOPE(0);
+
     perl_destruct(my_perl);
     perl_free(my_perl);
 
@@ -112,8 +147,8 @@ IV p5_sv_iv(PerlInterpreter *my_perl, SV* sv) {
     return SvIV(sv);
 }
 
-NV p5_sv_nv(PerlInterpreter *my_perl, SV* sv) {
-    return SvNV(sv);
+MYNV p5_sv_nv(PerlInterpreter *my_perl, SV* sv) {
+    return (MYNV) SvNV(sv);
 }
 
 SV *p5_sv_rv(PerlInterpreter *my_perl, SV* sv) {
@@ -133,7 +168,22 @@ int p5_is_array(PerlInterpreter *my_perl, SV* sv) {
 }
 
 int p5_is_hash(PerlInterpreter *my_perl, SV* sv) {
-    return (SvROK(sv) && SvTYPE(SvRV(sv)) == SVt_PVHV);
+    MAGIC *mg;
+    return (
+        (SvROK(sv) && SvTYPE(SvRV(sv)) == SVt_PVHV)
+        ? ((mg = mg_find(SvRV(sv), PERL_MAGIC_tied)) && sv_isa(mg->mg_obj, "Perl6::Hash"))
+            ? 2
+            : 1
+        : 0
+    );
+}
+
+IV p5_unwrap_p6_hash(PerlInterpreter *my_perl, SV *obj) {
+    MAGIC * const tie_mg = mg_find(SvRV(obj), PERL_MAGIC_tied);
+    SV * const hash = tie_mg->mg_obj;
+    SV * const p6hashobj = *(av_fetch((AV *) SvRV(hash), 0, 0));
+    MAGIC * const mg = mg_find(SvRV(p6hashobj), '~');
+    return ((_perl6_hash_magic*)(mg->mg_ptr))->index;
 }
 
 int p5_is_scalar_ref(PerlInterpreter *my_perl, SV* sv) {
@@ -180,8 +230,8 @@ SV *p5_int_to_sv(PerlInterpreter *my_perl, IV value) {
     return newSViv(value);
 }
 
-SV *p5_float_to_sv(PerlInterpreter *my_perl, NV value) {
-    return newSVnv(value);
+SV *p5_float_to_sv(PerlInterpreter *my_perl, MYNV value) {
+    return newSVnv((NV)value);
 }
 
 SV *p5_str_to_sv(PerlInterpreter *my_perl, STRLEN len, char* value) {
@@ -203,6 +253,13 @@ SV *p5_av_fetch(PerlInterpreter *my_perl, AV *av, I32 key) {
     return NULL;
 }
 
+void p5_av_store(PerlInterpreter *my_perl, AV *av, I32 key, SV *val) {
+    SvREFCNT_inc(val);
+    if (av_store(av, key, val) == NULL)
+        SvREFCNT_dec(val);
+    return;
+}
+
 void p5_av_push(PerlInterpreter *my_perl, AV *av, SV *sv) {
     av_push(av, sv);
 }
@@ -223,10 +280,21 @@ SV *p5_hv_iterval(PerlInterpreter *my_perl, HV *hv, HE *entry) {
     return hv_iterval(hv, entry);
 }
 
+SV *p5_hv_fetch(PerlInterpreter *my_perl, HV *hv, STRLEN len, const char *key) {
+    SV ** const item = hv_fetch(hv, key, len, 0);
+    if (item)
+        return *item;
+    return NULL;
+}
+
 void p5_hv_store(PerlInterpreter *my_perl, HV *hv, const char *key, SV *val) {
     hv_store(hv, key, strlen(key), val, 0);
 }
 
+int p5_hv_exists(PerlInterpreter *my_perl, HV *hv, STRLEN len, const char *key) {
+    return hv_exists(hv, key, len);
+}
+
 SV *p5_undef(PerlInterpreter *my_perl) {
     return &PL_sv_undef;
 }
@@ -243,10 +311,64 @@ SV *p5_newRV_noinc(PerlInterpreter *my_perl, SV *sv) {
     return newRV_noinc(sv);
 }
 
+SV *p5_newRV_inc(PerlInterpreter *my_perl, SV *sv) {
+    return newRV_inc(sv);
+}
+
 const char *p5_sv_reftype(PerlInterpreter *my_perl, SV *sv) {
     return sv_reftype(SvRV(sv), 1);
 }
 
+I32 p5_get_type(PerlInterpreter *my_perl, SV *sv) {
+    int is_hash;
+    if (p5_is_object(my_perl, sv)) {
+        return 1;
+    }
+    else if (p5_is_sub_ref(my_perl, sv)) {
+        return 2;
+    }
+    else if (p5_SvNOK(my_perl, sv)) {
+        return 3;
+    }
+    else if (p5_SvIOK(my_perl, sv)) {
+        return 4;
+    }
+    else if (p5_SvPOK(my_perl, sv)) {
+        return 5;
+    }
+    else if (p5_is_array(my_perl, sv)) {
+        return 6;
+    }
+    else if ((is_hash = p5_is_hash(my_perl, sv)) > 0) {
+        return 6 + is_hash;
+    }
+    else if (p5_is_undef(my_perl, sv)) {
+        return 9;
+    }
+    else if (p5_is_scalar_ref(my_perl, sv)) {
+        return 10;
+    }
+    else {
+        return 0;
+    }
+}
+
+SV *p5_get_global(PerlInterpreter *my_perl, const char* name) {
+    if (strlen(name) < 2)
+        return NULL;
+
+    if (name[0] == '$')
+        return get_sv(&name[1], 0);
+
+    if (name[0] == '@')
+        return sv_2mortal(newRV_inc((SV *)get_av(&name[1], 0)));
+
+    if (name[0] == '%')
+        return sv_2mortal(newRV_inc((SV *)get_hv(&name[1], 0)));
+
+    return NULL;
+}
+
 SV *p5_eval_pv(PerlInterpreter *my_perl, const char* p, I32 croak_on_error) {
     PERL_SET_CONTEXT(my_perl);
     return eval_pv(p, croak_on_error);
@@ -256,63 +378,125 @@ SV *p5_err_sv(PerlInterpreter *my_perl) {
     return sv_mortalcopy(ERRSV);
 }
 
-AV *p5_call_package_method(PerlInterpreter *my_perl, char *package, char *name, int len, SV *args[]) {
-    dSP;
-    int i;
-    I32 count;
-    AV * const retval = newAV();
-    int flags = G_ARRAY | G_EVAL;
-
+SV *p5_call_package_method(PerlInterpreter *my_perl, char *package, char *name, int len, SV *args[], I32 *count) {
     PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        int i;
+        SV * retval = NULL;
+        int flags = G_ARRAY | G_EVAL;
 
-    ENTER;
-    SAVETMPS;
+        ENTER;
+        SAVETMPS;
 
-    PUSHMARK(SP);
+        PUSHMARK(SP);
 
-    XPUSHs(newSVpv(package, 0));
-    for (i = 0; i < len; i++) {
-        XPUSHs(sv_2mortal(args[i]));
+        XPUSHs(newSVpv(package, 0));
+        for (i = 0; i < len; i++) {
+            XPUSHs(sv_2mortal(args[i]));
+        }
+
+        PUTBACK;
+
+        *count = call_method(name, flags);
+        SPAGAIN;
+
+        if (*count == 1) {
+            retval = POPs;
+            SvREFCNT_inc(retval);
+        }
+        else {
+            if (*count > 1) {
+                retval = (SV *)newAV();
+                av_extend((AV *)retval, *count - 1);
+            }
+
+            for (i = *count - 1; i >= 0; i--) {
+                SV * const next = POPs;
+                SvREFCNT_inc(next);
+
+                if (av_store((AV *)retval, i, next) == NULL)
+                    SvREFCNT_dec(next); /* see perlguts Working with AVs */
+            }
+        }
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return retval;
     }
-
-    PUTBACK;
-
-    count = call_method(name, flags);
-    SPAGAIN;
-
-    if (count > 0)
-        av_extend(retval, count - 1);
-
-    for (i = count - 1; i >= 0; i--) {
-        SV * const next = POPs;
-        SvREFCNT_inc(next);
-
-        if (av_store(retval, i, next) == NULL)
-            SvREFCNT_dec(next); /* see perlguts Working with AVs */
-    }
-
-    PUTBACK;
-    FREETMPS;
-    LEAVE;
-
-    return retval;
 }
 
-AV *p5_call_method(PerlInterpreter *my_perl, char *package, SV *obj, I32 context, char *name, int len, SV *args[]) {
-    dSP;
-    int i;
-    AV * const retval = newAV();
-    int flags = (context ? G_SCALAR : G_ARRAY) | G_EVAL;
-
+SV *p5_call_method(PerlInterpreter *my_perl, char *package, SV *obj, I32 context, char *name, int len, SV *args[], I32 *count) {
     PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        int i;
+        SV * retval = NULL;
+        int flags = (context ? G_SCALAR : G_ARRAY) | G_EVAL;
 
-    ENTER;
-    SAVETMPS;
+        ENTER;
+        SAVETMPS;
+
+        HV * const pkg = package != NULL ? gv_stashpv(package, 0) : SvSTASH((SV*)SvRV(obj));
+        GV * const gv = Perl_gv_fetchmethod_autoload(aTHX_ pkg, name, TRUE);
+        if (gv && isGV(gv)) {
+            PUSHMARK(SP);
+
+            for (i = 0; i < len; i++) {
+                XPUSHs(sv_2mortal(args[i]));
+            }
+
+            PUTBACK;
+
+            SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv)));
+
+            *count = call_sv(rv, flags);
+            SPAGAIN;
+
+            if (*count == 1) {
+                retval = POPs;
+                SvREFCNT_inc(retval);
+            }
+            else {
+                if (*count > 1) {
+                    retval = (SV *)newAV();
+                    av_extend((AV *)retval, *count - 1);
+                }
+                for (i = *count - 1; i >= 0; i--) {
+                    SV * const next = POPs;
+                    SvREFCNT_inc(next);
+
+                    if (av_store((AV *)retval, i, next) == NULL)
+                        SvREFCNT_dec(next); /* see perlguts Working with AVs */
+                }
+            }
+        }
+        else {
+            ERRSV = newSVpvf("Could not find method \"%s\" of \"%s\" object", name, HvNAME(pkg));
+        }
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return retval;
+    }
+}
+
+SV *p5_call_function(PerlInterpreter *my_perl, char *name, int len, SV *args[], I32 *count) {
+    PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        int i;
+        SV * retval = NULL;
+        int flags = G_ARRAY | G_EVAL;
+
+
+        ENTER;
+        SAVETMPS;
 
-    HV * const pkg = package != NULL ? gv_stashpv(package, 0) : SvSTASH((SV*)SvRV(obj));
-    GV * const gv = Perl_gv_fetchmethod_autoload(aTHX_ pkg, name, TRUE);
-    if (gv && isGV(gv)) {
-        I32 count;
         PUSHMARK(SP);
 
         for (i = 0; i < len; i++) {
@@ -321,123 +505,83 @@ AV *p5_call_method(PerlInterpreter *my_perl, char *package, SV *obj, I32 context
 
         PUTBACK;
 
-        SV * const rv = sv_2mortal(newRV((SV*)GvCV(gv)));
-
-        count = call_sv(rv, flags);
+        *count = call_pv(name, flags);
         SPAGAIN;
 
-        if (count > 0)
-            av_extend(retval, count - 1);
-        for (i = count - 1; i >= 0; i--) {
-            SV * const next = POPs;
-            SvREFCNT_inc(next);
-
-            if (av_store(retval, i, next) == NULL)
-                SvREFCNT_dec(next); /* see perlguts Working with AVs */
+        if (*count == 1) {
+            retval = POPs;
+            SvREFCNT_inc(retval);
         }
-    }
-    else {
-        ERRSV = newSVpvf("Could not find method \"%s\" of \"%s\" object", name, HvNAME(pkg));
-    }
+        else if (*count > 1) {
+            retval = (SV *)newAV();
+            av_extend((AV *)retval, *count - 1);
 
-    PUTBACK;
-    FREETMPS;
-    LEAVE;
+            for (i = *count - 1; i >= 0; i--) {
+                SV * const next = POPs;
+                SvREFCNT_inc(next);
 
-    return retval;
+                if (av_store((AV *)retval, i, next) == NULL)
+                    SvREFCNT_dec(next); /* see perlguts Working with AVs */
+            }
+        }
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return retval;
+    }
 }
 
-AV *p5_call_function(PerlInterpreter *my_perl, char *name, int len, SV *args[]) {
-    dSP;
-    int i;
-    I32 count;
-    AV * const retval = newAV();
-    int flags = G_ARRAY | G_EVAL;
-
+SV *p5_call_code_ref(PerlInterpreter *my_perl, SV *code_ref, int len, SV *args[], I32 *count) {
     PERL_SET_CONTEXT(my_perl);
+    {
+        dSP;
+        int i;
+        SV * retval;
+        int flags = G_ARRAY | G_EVAL;
 
-    ENTER;
-    SAVETMPS;
 
-    PUSHMARK(SP);
+        ENTER;
+        SAVETMPS;
 
-    for (i = 0; i < len; i++) {
-        XPUSHs(sv_2mortal(args[i]));
+        PUSHMARK(SP);
+
+        for (i = 0; i < len; i++) {
+            XPUSHs(sv_2mortal(args[i]));
+        }
+
+        PUTBACK;
+
+        *count = call_sv(code_ref, flags);
+        SPAGAIN;
+
+        if (*count == 1) {
+            retval = POPs;
+            SvREFCNT_inc(retval);
+        } else if (*count > 1) {
+            retval = (SV *)newAV();
+            av_extend((AV *)retval, *count - 1);
+
+            for (i = *count - 1; i >= 0; i--) {
+                SV * const next = POPs;
+                SvREFCNT_inc(next);
+
+                if (av_store((AV *)retval, i, next) == NULL)
+                    SvREFCNT_dec(next); /* see perlguts Working with AVs */
+            }
+        }
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return retval;
     }
-
-    PUTBACK;
-
-    count = call_pv(name, flags);
-    SPAGAIN;
-
-    if (count > 0)
-        av_extend(retval, count - 1);
-    for (i = count - 1; i >= 0; i--) {
-        SV * const next = POPs;
-        SvREFCNT_inc(next);
-
-        if (av_store(retval, i, next) == NULL)
-            SvREFCNT_dec(next); /* see perlguts Working with AVs */
-    }
-
-    PUTBACK;
-    FREETMPS;
-    LEAVE;
-
-    return retval;
 }
 
-AV *p5_call_code_ref(PerlInterpreter *my_perl, SV *code_ref, int len, SV *args[]) {
-    dSP;
-    int i;
-    I32 count;
-    AV * const retval = newAV();
-    int flags = G_ARRAY | G_EVAL;
-
-    PERL_SET_CONTEXT(my_perl);
-
-    ENTER;
-    SAVETMPS;
-
-    PUSHMARK(SP);
-
-    for (i = 0; i < len; i++) {
-        XPUSHs(sv_2mortal(args[i]));
-    }
-
-    PUTBACK;
-
-    count = call_sv(code_ref, flags);
-    SPAGAIN;
-
-    if (count > 0)
-        av_extend(retval, count - 1);
-    for (i = count - 1; i >= 0; i--) {
-        SV * const next = POPs;
-        SvREFCNT_inc(next);
-
-        if (av_store(retval, i, next) == NULL)
-            SvREFCNT_dec(next); /* see perlguts Working with AVs */
-    }
-
-    PUTBACK;
-    FREETMPS;
-    LEAVE;
-
-    return retval;
-}
-
-typedef struct {
-    I32 key; /* to make sure it came from Inline */
-    IV index;
-    union {
-        SV *(*call_p6_method)(IV, char *, I32, SV *, SV **);
-        SV *(*call_p6_callable)(IV, SV *, SV **);
-    };
-    void (*free_p6_object)(IV);
-} _perl6_magic;
-
 #define PERL6_MAGIC_KEY 0x0DD515FE
+#define PERL6_HASH_MAGIC_KEY 0x0DD515FF
 
 int p5_free_perl6_obj(pTHX_ SV* obj, MAGIC *mg)
 {
@@ -448,6 +592,15 @@ int p5_free_perl6_obj(pTHX_ SV* obj, MAGIC *mg)
     return 0;
 }
 
+int p5_free_perl6_hash(pTHX_ SV* obj, MAGIC *mg)
+{
+    if (mg) {
+        _perl6_hash_magic* const p6mg = (_perl6_hash_magic*) mg->mg_ptr;
+        p6mg->free_p6_object(p6mg->index);
+    }
+    return 0;
+}
+
 MGVTBL p5_inline_mg_vtbl = {
     0x0,
     0x0,
@@ -459,6 +612,17 @@ MGVTBL p5_inline_mg_vtbl = {
     0x0
 };
 
+MGVTBL p5_inline_hash_mg_vtbl = {
+    0x0,
+    0x0,
+    0x0,
+    0x0,
+    &p5_free_perl6_hash,
+    0x0,
+    0x0,
+    0x0
+};
+
 void p5_rebless_object(PerlInterpreter *my_perl, SV *obj, char *package, IV i, SV *(*call_p6_method)(IV, char * , I32, SV *, SV **), void (*free_p6_object)(IV)) {
     SV * const inst = SvRV(obj);
     HV *stash = gv_stashpv(package, GV_ADD);
@@ -504,9 +668,25 @@ SV *p5_wrap_p6_object(PerlInterpreter *my_perl, IV i, SV *p5obj, SV *(*call_p6_m
 SV *p5_wrap_p6_callable(PerlInterpreter *my_perl, IV i, SV *p5obj, SV *(*call)(IV, SV *, SV **), void (*free_p6_object)(IV)) {
     SV * inst;
     SV * inst_ptr;
+
+    PERL_SET_CONTEXT(my_perl);
+
     if (p5obj == NULL) {
-        inst_ptr = newSViv(0);
-        inst = newSVrv(inst_ptr, "Perl6::Callable");
+        dSP;
+        ENTER;
+        SAVETMPS;
+
+        PUSHMARK(SP);
+        call_pv("Perl6::Callable::new", G_SCALAR);
+        SPAGAIN;
+
+        inst_ptr = POPs;
+        inst = SvRV(inst_ptr);
+        SvREFCNT_inc(inst_ptr);
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
     }
     else {
         inst_ptr = p5obj;
@@ -527,34 +707,87 @@ SV *p5_wrap_p6_callable(PerlInterpreter *my_perl, IV i, SV *p5obj, SV *(*call)(I
     return inst_ptr;
 }
 
-SV *p5_wrap_p6_handle(PerlInterpreter *my_perl, IV i, SV *p5obj, SV *(*call_p6_method)(IV, char * , I32, SV *, SV **), void (*free_p6_object)(IV)) {
-    SV *handle = p5_wrap_p6_object(my_perl, i, p5obj, call_p6_method, free_p6_object);
-    int flags = G_SCALAR;
-    dSP;
-
+SV *p5_wrap_p6_hash(
+    PerlInterpreter *my_perl,
+    IV i,
+    SV *(*call_p6_method)(IV, char * , I32, SV *, SV **),
+    SV *(*hash_at_key)(IV, char *),
+    SV *(*hash_assign_key)(IV, char *, SV *),
+    void (*free_p6_object)(IV)
+) {
     PERL_SET_CONTEXT(my_perl);
+    {
+        int flags = G_SCALAR;
+        dSP;
 
-    ENTER;
-    SAVETMPS;
+        SV * inst;
+        SV * inst_ptr;
+        inst_ptr = newSViv(0); // will be upgraded to an RV
+        inst = newSVrv(inst_ptr, "Perl6::Object");
+        _perl6_hash_magic priv;
 
-    PUSHMARK(SP);
+        /* set up magic */
+        priv.key = PERL6_HASH_MAGIC_KEY;
+        priv.index = i;
+        priv.call_p6_method  = call_p6_method;
+        priv.hash_at_key     = hash_at_key;
+        priv.hash_assign_key = hash_assign_key;
+        priv.free_p6_object  = free_p6_object;
+        sv_magicext(inst, inst, PERL_MAGIC_ext, &p5_inline_hash_mg_vtbl, (char *) &priv, sizeof(priv));
 
-    XPUSHs(newSVpv("Perl6::Handle", 0));
-    XPUSHs(handle);
+        ENTER;
+        SAVETMPS;
 
-    PUTBACK;
+        PUSHMARK(SP);
 
-    call_method("new", flags);
-    SPAGAIN;
+        XPUSHs(newSVpv("Perl6::Hash", 0));
+        XPUSHs(inst_ptr);
 
-    SV *tied_handle = POPs;
-    SvREFCNT_inc(tied_handle);
+        PUTBACK;
 
-    PUTBACK;
-    FREETMPS;
-    LEAVE;
+        call_method("new", flags);
+        SPAGAIN;
 
-    return tied_handle;
+        SV *tied_handle = POPs;
+        SvREFCNT_inc(tied_handle);
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return tied_handle;
+    }
+}
+
+SV *p5_wrap_p6_handle(PerlInterpreter *my_perl, IV i, SV *p5obj, SV *(*call_p6_method)(IV, char * , I32, SV *, SV **), void (*free_p6_object)(IV)) {
+    PERL_SET_CONTEXT(my_perl);
+    {
+        SV *handle = p5_wrap_p6_object(my_perl, i, p5obj, call_p6_method, free_p6_object);
+        int flags = G_SCALAR;
+        dSP;
+
+        ENTER;
+        SAVETMPS;
+
+        PUSHMARK(SP);
+
+        XPUSHs(newSVpv("Perl6::Handle", 0));
+        XPUSHs(handle);
+
+        PUTBACK;
+
+        call_method("new", flags);
+        SPAGAIN;
+
+        SV *tied_handle = POPs;
+        SvREFCNT_inc(tied_handle);
+
+        PUTBACK;
+        FREETMPS;
+        LEAVE;
+
+        return tied_handle;
+    }
 }
 
 int p5_is_wrapped_p6_object(PerlInterpreter *my_perl, SV *obj) {
@@ -624,6 +857,47 @@ XS(p5_call_p6_method) {
     }
 }
 
+XS(p5_hash_at_key) {
+    dXSARGS;
+    SV * self = ST(0);
+    SV * key  = ST(1);
+
+    SV * const p6hashobj = *(av_fetch((AV *) SvRV(self), 0, 0));
+    MAGIC * const mg = mg_find(SvRV(p6hashobj), '~');
+    _perl6_hash_magic* const p6mg = (_perl6_hash_magic*)(mg->mg_ptr);
+
+    STRLEN len;
+    char * const key_pv  = SvPV(key, len);
+
+    SV * retval = p6mg->hash_at_key(p6mg->index, key_pv);
+
+    sv_2mortal(retval);
+    sp -= items;
+
+    XPUSHs(retval);
+    XSRETURN(1);
+}
+
+XS(p5_hash_assign_key) {
+    dXSARGS;
+    SV * self = ST(0);
+    SV * key  = ST(1);
+    SV * val  = ST(2);
+
+    SV * const p6hashobj = *(av_fetch((AV *) SvRV(self), 0, 0));
+    MAGIC * const mg = mg_find(SvRV(p6hashobj), '~');
+    _perl6_hash_magic* const p6mg = (_perl6_hash_magic*)(mg->mg_ptr);
+
+    STRLEN len;
+    char * const key_pv  = SvPV(key, len);
+
+    p6mg->hash_assign_key(p6mg->index, key_pv, val);
+
+    sp -= items;
+
+    XSRETURN_EMPTY;
+}
+
 XS(p5_call_p6_callable) {
     dXSARGS;
     SV * obj = ST(0);
